from typing import Self
from scipy.special import softmax
from scipy.stats import multivariate_normal
import numpy as np

from luma.core.super import Estimator, Evaluator, Supervised
from luma.interface.util import Matrix, Scalar, Vector, KernelUtil
from luma.interface.exception import NotFittedError
from luma.metric.classification import Accuracy


__all__ = ("LDAClassifier", "QDAClassifier", "RDAClassifier", "KDAClassifier")


class LDAClassifier(Estimator, Supervised):
    """
    Linear Discriminant Analysis (LDA) is a statistical technique used for
    dimensionality reduction and classification. It projects data onto a
    lower-dimensional space with the goal of maximizing class separability.
    LDA assumes that different classes generate data based on Gaussian
    distributions with the same covariance matrix but different mean vectors.
    It finds linear combinations of features that best separate the classes,
    facilitating classification or visualization of complex data.

    Notes
    -----
    * To use LDA for dimensionality reduction, refer to
        `luma.reduction.linear.LDA`
    """

    def __init__(self) -> None:
        self.means = (None,)
        self.priors = (None,)
        self.covs = None
        self._fitted = False

    def fit(self, X: Matrix, y: Vector) -> Self:
        m, n = X.shape
        class_labels = np.unique(y)
        n_classes = len(class_labels)

        self.means = np.zeros((n_classes, n))
        self.priors = np.zeros(n_classes)
        self.covs = np.zeros((n, n))

        for i, cl in enumerate(class_labels):
            X_class = X[y == cl]
            self.means[i] = X_class.mean(axis=0)
            self.priors[i] = X_class.shape[0] / m
            self.covs += np.cov(X_class, rowvar=False) * (X_class.shape[0] - 1)

        self.covs /= X.shape[0] - n_classes
        self.cov_inv = np.linalg.inv(self.covs)

        self.coef_ = np.dot(self.cov_inv, self.means.T).T
        self.intercept_ = -0.5 * np.diag(np.dot(self.means, self.coef_.T))
        self.intercept_ += np.log(self.priors)

        self._fitted = True
        return self

    def predict(self, X: Matrix) -> Vector:
        if not self._fitted:
            raise not NotFittedError(self)
        scores = np.dot(X, self.coef_.T) + self.intercept_

        return np.argmax(scores, axis=1)

    def predict_proba(self, X: Matrix) -> Matrix:
        if not self._fitted:
            raise NotFittedError(self)
        scores = np.dot(X, self.coef_.T) + self.intercept_

        return softmax(scores, axis=1)

    def score(self, X: Matrix, y: Matrix, metric: Evaluator = Accuracy) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)


class QDAClassifier(Estimator, Supervised):
    """
    Quadratic Discriminant Analysis (QDA) is a classification technique
    that assumes data from each class follows a Gaussian distribution
    with class-specific covariance matrices. It calculates the probability
    of a given sample belonging to each class based on a quadratic decision
    boundary determined by these distributions. Unlike Linear Discriminant
    Analysis (LDA), QDA allows for the modeling of nonlinear relationships
    due to the class-specific covariances. It is well-suited for datasets
    where classes exhibit different levels of variance.
    """

    def __init__(self) -> None:
        self.means = None
        self.covs = None
        self.priors = None
        self.classes = None
        self._fitted = False

    def fit(self, X: Matrix, y: Vector) -> Self:
        m, n = X.shape
        self.classes = np.unique(y)
        n_classes = len(self.classes)

        self.means = np.zeros((n_classes, n))
        self.covs = np.zeros((n_classes, n, n))
        self.priors = np.zeros(n_classes)

        for i, cl in enumerate(self.classes):
            X_cls = X[y == cl]
            self.means[i] = np.mean(X_cls, axis=0)
            self.covs[i] = np.cov(X_cls, rowvar=False)
            self.priors[i] = X_cls.shape[0] / m

        self._fitted = True
        return self

    def predict(self, X: Matrix) -> Vector:
        if not self._fitted:
            raise not NotFittedError(self)
        y_pred = [self._predict_sample(x) for x in X]
        return Vector(y_pred)

    def _predict_sample(self, x: Vector) -> Scalar:
        discs = np.zeros(len(self.classes))
        for i in range(len(self.classes)):
            diff = x - self.means[i]
            cov_inv = np.linalg.inv(self.covs[i])
            cov_det = np.linalg.det(self.covs[i])

            disc = np.log(cov_det) + diff.T.dot(cov_inv).dot(diff)
            disc *= -0.5
            disc += np.log(self.priors[i])
            discs[i] = disc

        return self.classes[np.argmax(discs)]

    def predict_proba(self, X: Matrix) -> Matrix:
        if not self._fitted:
            raise NotFittedError(self)
        log_probs = np.array([self._log_proba_sample(x) for x in X])

        return softmax(log_probs, axis=1)

    def _log_proba_sample(self, x: Vector) -> Vector:
        log_probs = np.zeros(len(self.classes))
        for i, _ in enumerate(self.classes):
            mvn = multivariate_normal(mean=self.means[i], cov=self.covs[i])
            log_probs[i] = mvn.logpdf(x) + np.log(self.priors[i])

        return log_probs

    def score(self, X: Matrix, y: Matrix, metric: Evaluator = Accuracy) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)


class RDAClassifier(Estimator, Supervised):
    """
    Regularized Discriminant Analysis (RDA) combines the features of
    Linear Discriminant Analysis (LDA) and Quadratic Discriminant Analysis
    (QDA) by introducing regularization to the covariance matrices. It
    adjusts the covariance estimates with a blend between the pooled
    covariance (LDA approach) and class-specific covariances (QDA approach),
    enhancing classification performance. RDA is particularly effective in
    scenarios with high-dimensional data or when the number of samples is
    limited. The method employs regularization parameters to balance bias
    and variance, aiming to optimize model accuracy.

    Parameters
    ----------
    `alpha` : Balancing parameter between the class-specific covariance matrices
    (0 for LDA-like, 1 for QDA-like approach)
    `gamma` : Shrinkage applied to the covariance matrices
    (0 for large shrinkage, i.e. max regularization)

    """

    def __init__(self, alpha: float = 0.5, gamma: float = 0.5) -> None:
        self.alpha = alpha
        self.gamma = gamma
        self.classes = None
        self.means = None
        self.priors = None
        self.covs = None
        self._fitted = False

        self.set_param_ranges({"alpha": ("0,+1", None), "gamma": ("0<,+inf", None)})
        self.check_param_ranges()

    def fit(self, X: Matrix, y: Vector) -> Self:
        m, n = X.shape
        self.classes = np.unique(y)
        n_classes = len(self.classes)

        self.means = np.zeros((n_classes, n))
        self.pooled_cov = np.zeros((n, n))
        self.covs = np.zeros((n_classes, n, n))
        self.priors = np.zeros(n_classes)

        for i, cl in enumerate(self.classes):
            X_cls = X[y == cl]
            self.means[i] = np.mean(X_cls, axis=0)
            self.priors[i] = X_cls.shape[0] / m
            class_cov = np.cov(X_cls, rowvar=False)
            self.pooled_cov += class_cov * (X_cls.shape[0] - 1)

        self.pooled_cov /= m - n_classes

        for i in range(n_classes):
            X_cls = X[y == self.classes[i]]
            class_cov = np.cov(X_cls, rowvar=False)
            self.covs[i] = self.alpha * class_cov
            self.covs[i] += (1 - self.alpha) * self.pooled_cov

            self.covs[i] = self.gamma * self.covs[i]
            self.covs[i] += (1 - self.gamma) * np.eye(n) * np.trace(self.covs[i])
            self.covs[i] /= n

        self._fitted = True
        return self

    def predict(self, X: Matrix) -> Vector:
        if not self._fitted:
            raise not NotFittedError(self)
        return Vector([self._predict_sample(x) for x in X])

    def _predict_sample(self, x: Vector) -> Scalar:
        discs = np.zeros(len(self.classes))
        for i in range(len(self.classes)):
            diff = x - self.means[i]
            cov_inv = np.linalg.inv(self.covs[i])
            cov_det = np.linalg.det(self.covs[i])

            disc = np.log(cov_det) + diff.T.dot(cov_inv).dot(diff)
            disc *= -0.5
            disc += np.log(self.priors[i])
            discs[i] = disc

        return self.classes[np.argmax(discs)]

    def predict_proba(self, X: Matrix) -> Matrix:
        if not self._fitted:
            raise NotFittedError(self)
        log_probs = np.array([self._log_proba_sample(x) for x in X])

        return softmax(log_probs, axis=1)

    def _log_proba_sample(self, x: Vector) -> Vector:
        log_probs = np.zeros(len(self.classes))
        for i, _ in enumerate(self.classes):
            mvn = multivariate_normal(mean=self.means[i], cov=self.covs[i])
            log_probs[i] = mvn.logpdf(x) + np.log(self.priors[i])

        return log_probs

    def score(self, X: Matrix, y: Matrix, metric: Evaluator = Accuracy) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)


class KDAClassifier(Estimator, Supervised):
    """
    A Kernel Discriminant Analysis (KDA) classifier is an extension of
    traditional discriminant analysis methods that incorporates kernel
    techniques to perform classification tasks in a transformed,
    high-dimensional feature space. By applying a kernel function, it
    enables the classifier to find nonlinear boundaries between classes,
    enhancing its ability to handle complex patterns that are not linearly
    separable. The KDA classifier operates by projecting input data into
    a space where classes are maximally distant from each other, using
    the calculated projections to classify new instances based on their
    proximity to class centroids.

    Parameters
    ----------
    `deg` : Polynomial degree of `poly` kernel
    `gamma` : Shape parameter of `rbf`, `sigmoid`, `laplacian`
    `coef` : Additional coefficient of `poly`, `sigmoid`
    `kernel` : Type of kernel functions

    Notes
    -----
    * To use KDA for dimensionality reduction, refer to
        `luma.reduction.linear.KDA`

    """

    def __init__(
        self,
        deg: int = 2,
        gamma: float = 1.0,
        coef: int = 0.0,
        kernel: KernelUtil.FuncType = "rbf",
    ) -> None:
        self.deg = deg
        self.gamma = gamma
        self.coef = coef
        self.kernel = kernel
        self.X_ = None
        self._fitted = False

        self.kernel_params = {
            "deg": self.deg,
            "alpha": self.alpha,
            "gamma": self.gamma,
            "coef": self.coef,
        }

        self.set_param_ranges({"deg": ("0,+inf", int), "gamma": ("0<,+inf", None)})
        self.check_param_ranges()

    def fit(self, X: Matrix, y: Vector) -> Self:
        m, _ = X.shape
        self._X = X
        self.classes = np.unique(y)
        self.n_classes = len(self.classes)

        self.ku_ = KernelUtil(self.kernel, **self.kernel_params)
        self.class_means = np.zeros((self.n_classes, m))

        self.K = self.ku_.kernel_func(X)
        for i, cl in enumerate(self.classes):
            self.class_means[i] = np.mean(self.K[y == cl], axis=0)

        self._fitted = False
        return self

    def _project(self, X: Matrix) -> Vector:
        K = self.ku_.kernel_func(X, self._X)
        proj = np.dot(K, self.class_means.T)
        return proj

    def predict(self, X: Matrix) -> Vector:
        if not self._fitted:
            raise NotFittedError(self)
        proj = self._project(X)

        return self.classes[np.argmax(proj, axis=1)]

    def predict_proba(self, X: Matrix) -> Matrix:
        if not self._fitted:
            raise NotFittedError(self)
        proj = self._project(X)

        return softmax(proj, axis=1)

    def score(self, X: Matrix, y: Matrix, metric: Evaluator = Accuracy) -> float:
        X_pred = self.predict(X)
        return metric.score(y_true=y, y_pred=X_pred)



from itertools import combinations
from typing import Tuple
import numpy as np

from luma.core.super import Estimator, Evaluator, Transformer, Supervised
from luma.interface.util import Matrix, Vector, Clone
from luma.interface.exception import NotFittedError, UnsupportedParameterError
from luma.model_selection.split import TrainTestSplit
from luma.model_selection.cv import CrossValidator
from luma.model_selection.fold import FoldType, KFold


__all__ = ("SBS", "SFS", "RFE")


class SBS(Transformer, Transformer.Feature, Supervised):
    """
    Sequential Backward Selection (SBS) is a feature selection technique
    used in machine learning. It starts with all features and iteratively
    removes the least significant feature at each step_size. The goal is to
    reduce dimensionality while maintaining or improving model performance.
    The process continues until the desired number of features is reached
    or performance criteria are met.

    Parameters
    ----------
    `estimator` : An estimator to fit and evaluate
    `n_features` : Number of features to select (`0~1` value for proportion)
    `metric` : Scoring metric for selecting features
    `test_size` : Proportional size of the validation set
    `cv` : K-fold size for cross validation (`0` to disable CV)
    `shuffle` : Whether to shuffle the dataset
    `stratify` : Whether to perform stratified split
    `fold_type` : Fold type (Default `KFold`)
    `random_state` : Seed for splitting the data

    Notes
    -----
    * An instance of the estimator must be passed to `estimator`
    * For `metric`, both class or instance are possible

    Examples
    --------
    >>> sbs = SBS(estimator=AnyEstimator(),
                  n_features=0.25,
                  metric=AnyEvaluator,
                  test_size=0.2,
                  cv=5,
                  random_state=None)
    >>> sbs.fit(X, y)
    >>> Z = sbs.transform(X)

    """

    def __init__(
        self,
        estimator: Estimator = None,
        n_features: int | float = 1,
        metric: Evaluator = None,
        test_size: float = 0.2,
        cv: int = 5,
        shuffle: bool = True,
        stratify: bool = False,
        fold_type: FoldType = KFold,
        random_state: int = None,
        verbose: bool = False,
    ) -> None:
        self.estimator = estimator
        self.n_features = n_features
        self.metric = metric
        self.test_size = test_size
        self.cv = cv
        self.shuffle = shuffle
        self.stratify = stratify
        self.fold_type = fold_type
        self.random_state = random_state
        self.verbose = verbose
        self._fitted = False

        self.set_param_ranges(
            {
                "n_features": ("0<,+inf", None),
                "test_size": ("0,1", float),
                "cv": ("0<,+inf", int),
            }
        )
        self.check_param_ranges()

    def fit(self, X: Matrix, y: Vector) -> "SBS":
        _, n = X.shape
        self.estimator = Clone(self.estimator).get

        if 0 < self.n_features < 1:
            self.n_features = np.ceil(n * self.n_features).astype(int)
        elif self.n_features <= 0 or self.n_features > n:
            raise UnsupportedParameterError(self.n_features)

        if self.cv:
            cv_model = CrossValidator(
                estimator=self.estimator,
                metric=self.metric,
                cv=self.cv,
                shuffle=self.shuffle,
                fold_type=self.fold_type,
                random_state=self.random_state,
            )
        else:
            Xy_split = TrainTestSplit(
                X,
                y,
                test_size=self.test_size,
                shuffle=self.shuffle,
                stratify=self.stratify,
                random_state=self.random_state,
            ).get
            X_train, X_test, y_train, y_test = Xy_split

        self.indices = tuple(range(n))
        self.subsets = [self.indices]

        if self.cv:
            _, score = cv_model.score(X, y)
        else:
            score = self._calculate_score(
                X_train, X_test, y_train, y_test, indices=self.indices
            )
        self.scores = [score]

        iter = 1
        while n > self.n_features:
            scores = []
            subsets = []
            for p in combinations(self.indices, n - 1):
                if self.cv:
                    _, score = cv_model.score(X[:, p], y)
                else:
                    score = self._calculate_score(
                        X_train, X_test, y_train, y_test, indices=p
                    )
                scores.append(score)
                subsets.append(p)

            best = np.argmax(scores)
            self.indices = subsets[best]
            self.subsets.append(self.indices)
            self.scores.append(scores[best])

            if self.verbose:
                print(
                    f"[SBS] Feature added: {self.indices[-1]}",
                    f"with score: {scores[best]}",
                )

            n -= 1
            iter += 1

        self.n_score = self.scores[-1]
        if self.verbose:
            print(f"[SBS] Selection finished with final features: {self.indices}")

        self._fitted = True
        return self

    def _calculate_score(
        self,
        X_train: Matrix,
        X_test: Matrix,
        y_train: Matrix,
        y_test: Matrix,
        indices: Tuple,
    ) -> float:
        self.estimator.fit(X_train[:, indices], y_train)
        y_pred = self.estimator.predict(X_test[:, indices])

        return self.metric.score(y_true=y_test, y_pred=y_pred)

    def transform(self, X: Matrix) -> Matrix:
        if not self._fitted:
            raise NotFittedError(self)
        return X[:, self.indices]

    def fit_transform(self, X: Matrix, y: Vector) -> Matrix:
        self.fit(X, y)
        return self.transform(X)


class SFS(Transformer, Transformer.Feature, Supervised):
    """
    Sequential Forward Selection (SFS) is a feature selection technique
    used in machine learning. It starts with no features and iteratively
    adds the most significant feature at each step. The goal is to
    reduce dimensionality while maintaining or improving model performance.
    The process continues until the desired number of features is reached
    or performance criteria are met.

    Parameters
    ----------
    `estimator` : An estimator to fit and evaluate
    `n_features` : Number of features to select (`0~1` value for proportion)
    `metric` : Scoring metric for selecting features
    `test_size` : Proportional size of the validation set
    `cv` : K-fold size for cross validation (`0` to disable CV)
    `shuffle` : Whether to shuffle the dataset
    `stratify` : Whether to perform stratified split
    `fold_type` : Fold type (Default `KFold`)
    `random_state` : Seed for splitting the data

    Notes
    -----
    * An instance of the estimator must be passed to `estimator`
    * Do not use the estimator for `SFS` if it is supposed to be the
        main estimator for the model
    * For `metric`, both class or instance are possible

    Examples
    --------
    >>> sfs = SFS(estimator=AnyEstimator(),
                  n_features=0.25,
                  metric=AnyEvaluator,
                  test_size=0.2,
                  cv=5,
                  random_state=None)
    >>> sfs.fit(X, y)
    >>> Z = sbs.transform(X)

    """

    def __init__(
        self,
        estimator: Estimator = None,
        n_features: int | float = 1,
        metric: Evaluator = None,
        test_size: float = 0.2,
        cv: int = 5,
        shuffle: bool = True,
        stratify: bool = False,
        fold_type: FoldType = KFold,
        random_state: int = None,
        verbose: bool = False,
    ) -> None:
        self.estimator = estimator
        self.n_features = n_features
        self.metric = metric
        self.test_size = test_size
        self.cv = cv
        self.shuffle = shuffle
        self.stratify = stratify
        self.fold_type = fold_type
        self.random_state = random_state
        self.verbose = verbose
        self._fitted = False

        self.set_param_ranges(
            {
                "n_features": ("0<,+inf", None),
                "test_size": ("0,1", float),
                "cv": ("0<,+inf", int),
            }
        )
        self.check_param_ranges()

    def fit(self, X: Matrix, y: Vector) -> "SFS":
        _, n = X.shape
        self.estimator = Clone(self.estimator).get

        if isinstance(self.n_features, float) and 0 < self.n_features < 1:
            self.n_features = np.ceil(n * self.n_features).astype(int)
        elif self.n_features <= 0 or self.n_features > n:
            raise UnsupportedParameterError(self.n_features)

        if self.cv:
            cv_model = CrossValidator(
                estimator=self.estimator,
                metric=self.metric,
                cv=self.cv,
                shuffle=self.shuffle,
                fold_type=self.fold_type,
                random_state=self.random_state,
            )
        else:
            Xy_split = TrainTestSplit(
                X,
                y,
                test_size=self.test_size,
                shuffle=self.shuffle,
                stratify=self.stratify,
                random_state=self.random_state,
            ).get
            X_train, X_test, y_train, y_test = Xy_split

        self.indices = ()
        self.subsets = [self.indices]
        score = -np.inf
        self.scores = [score]

        while len(self.indices) < self.n_features:
            scores = []
            subsets = []
            for p in combinations(range(n), len(self.indices) + 1):
                if set(self.indices).issubset(set(p)):
                    if self.cv:
                        _, score = cv_model.score(X[:, p], y)
                    else:
                        score = self._calculate_score(
                            X_train, X_test, y_train, y_test, indices=p
                        )
                    scores.append(score)
                    subsets.append(p)

            best = np.argmax(scores)
            self.indices = subsets[best]
            self.subsets.append(self.indices)
            self.scores.append(scores[best])

            if self.verbose:
                print(
                    f"[SFS] Feature added: {self.indices[-1]}",
                    f"with score: {scores[best]}",
                )

        self.n_score = self.scores[-1]
        if self.verbose:
            print(f"[SFS] Selection finished with final features: {self.indices}")

        self._fitted = True
        return self

    def _calculate_score(
        self,
        X_train: Matrix,
        X_test: Matrix,
        y_train: Matrix,
        y_test: Matrix,
        indices: Tuple,
    ) -> float:
        self.estimator.fit(X_train[:, indices], y_train)
        y_pred = self.estimator.predict(X_test[:, indices])

        return self.metric.score(y_true=y_test, y_pred=y_pred)

    def transform(self, X: Matrix) -> Matrix:
        if not self._fitted:
            raise NotFittedError(self)
        return X[:, self.indices]

    def fit_transform(self, X: Matrix, y: Vector) -> Matrix:
        self.fit(X, y)
        return self.transform(X)


class RFE(Transformer, Transformer.Feature, Supervised):
    """
    Recursive Feature Elimination (RFE) is a feature selection method
    used in machine learning. It works by recursively removing the least
    important features based on a given model's feature importance or
    coefficients. RFE starts with all features and eliminates a specified
    number or fraction of features at each step. The process continues
    until the desired number of features is reached. This method helps in
    enhancing model performance and interpretability by reducing complexity.

    Parameters
    ----------
    `estimator` : An estimator to fit and evaluate
    `n_features` : Number of features to select (`0~1` value for proportion)
    `metric` : Scoring metric for selecting features
    `step_size` : Number of features to eliminate in each step
    `cv` : K-fold size for cross validation (`0` to disable CV)
    `shuffle` : Whether to shuffle the dataset
    `fold_type` : Fold type (Default `KFold`)
    `random_state` : Seed for splitting the data

    Notes
    -----
    * An instance of the estimator must be passed to `estimator`
    * Do not use the estimator for `RFE` if it is supposed to be the
        main estimator for the model
    * For `metric`, both class or instance are possible

    Examples
    --------
    >>> rfe = RFE(estimator=AnyEstimator(),
                  n_features=0.25,
                  metric=AnyEvaluator,
                  step_size=1,
                  cv=5,
                  random_state=None)
    >>> rfe.fit(X, y)
    >>> Z = rfe.transform(X)

    """

    def __init__(
        self,
        estimator: Estimator = None,
        n_features: int | float = 1,
        metric: Evaluator = None,
        step_size: int = 1,
        cv: int = 5,
        shuffle: bool = True,
        fold_type: FoldType = KFold,
        random_state: int = None,
        verbose: bool = False,
    ) -> None:
        self.estimator = estimator
        self.n_features = n_features
        self.metric = metric
        self.step_size = step_size
        self.cv = cv
        self.shuffle = shuffle
        self.fold_type = fold_type
        self.random_state = random_state
        self.verbose = verbose
        self._fitted = False

        self.set_param_ranges(
            {
                "n_features": ("0<,+inf", None),
                "step_size": ("0<,+inf", int),
                "cv": ("0<,+inf", int),
            }
        )
        self.check_param_ranges()

    def fit(self, X: Matrix, y: Vector) -> "RFE":
        _, n = X.shape
        self.estimator = Clone(self.estimator).get

        if 0 < self.n_features < 1:
            self.n_features = np.floor(n * self.n_features).astype(int)
        elif self.n_features <= 0 or self.n_features > n:
            raise UnsupportedParameterError(self.n_features)

        self.cv_model = CrossValidator(
            estimator=self.estimator,
            metric=self.metric,
            cv=self.cv,
            shuffle=self.shuffle,
            fold_type=self.fold_type,
            random_state=self.random_state,
        )

        self.support = np.ones(n, dtype=bool)
        self.ranking = np.ones(n, dtype=int)

        while np.sum(self.support) > self.n_features:
            scores = self._calculate_score(X, y)
            indices = np.argsort(scores)[::-1][: self.step_size]
            self.support[indices] = False
            self.ranking[indices] += 1

            if self.verbose:
                print(
                    f"[RFE] Removed {self.step_size} features,",
                    f"remaining features: {np.sum(self.support)}",
                    f"with best-score: {scores[:self.step_size]}",
                )

        features = tuple(i for i, sup in enumerate(self.support) if sup)
        if self.verbose:
            print(f"[RFE] Elimination finished with final features: {features}")

        self._fitted = True
        return self

    def _calculate_score(self, X: Matrix, y: Vector) -> Vector[float]:
        scores = np.zeros(X.shape[1])
        for i in range(X.shape[1]):
            if not self.support[i]:
                continue
            indices = self.support.copy()
            indices[i] = False

            if self.cv:
                _, score = self.cv_model.score(X[:, indices], y)
            else:
                self.estimator.fit(X[:, indices], y)
                y_pred = self.estimator.predict(X[:, indices])
                score = self.metric.score(y_true=y, y_pred=y_pred)

            scores[i] = score

        return scores

    def transform(self, X: Matrix) -> Matrix:
        if not self._fitted:
            raise NotFittedError(self)
        return X[:, self.support]

    def fit_transform(self, X: Matrix, y: Vector) -> Matrix:
        self.fit(X, y)
        return self.transform(X)
